{--------------------------------------------------------------------
 EF_EDIT - Labeled input controls for Borland's Delphi product.
 v. 1.07a February 6, 1997 (for Delphi 1 and 2)
 Affordable Shareware; costs only U.S. $5 (together with EF_DBEDT)
---------------------------------------------------------------------
 Copyright  1996 Eric G.V. Fookes
 - All Rights Reserved -
 Address: Av. Eugne-Pittard 22Ter, CH - 1206 Geneva, Switzerland
 E-mail: fookes@sc2a.unige.ch
 Web: http://www.unige.ch/sciences/terre/geologie/fookes/delphi.htm
---------------------------------------------------------------------}
unit Ef_Edit;

interface

uses
  WinTypes, WinProcs, Classes, Graphics, Messages, Controls, StdCtrls, SysUtils,
  Mask, Printers;

type
  TLabelPos = (lpCustom,lpTop,lpLeft,lpBottom,lpRight);

const
  {Error ID for ELabeledControl exception}
  ElcDiffParent = 1;
  ElcDiffOwner  = 2;
  {The following constants make it easy to change the default label position}
  DefaultLabelPos = lpLeft;
  DefaultMargin = 8;
  {Defines the default value for the AddColon property}
  DefaultAddColon = True;
  {Height offset in pixels of label top from control window top
   when LabelPos is lpLeft or lpRight. A value of 3 works well under many resolutions.}
  AdjLabelToControl : Integer = 3;
  {Prefix used when (re-)naming attached labels}
  LabelPrefix : string[5] = 'el_';

type
  TNameChangeEvent = procedure (Sender: TObject; const NewName: TComponentName;
                                var AllowChange:Boolean) of object;

  TEnhMemo = class(TMemo)
    public
     procedure SetTabStops(Tabs: array of Word);
     procedure Copy2Printer;
     procedure ResetUndoFlag;
     function  CanUndo:Boolean;
     procedure UndoEdit;
     procedure CurrentCursorPos(var Row,Col:Integer);
     function  CursorRow:Integer;
     function  CursorCol:Integer;
  end;

  TEnhLabel = class(TLabel)
    protected
     FDisabled:Boolean;
     FDisabledColor,FKeepColor:TColor;
     FOnNewBounds:TNotifyEvent;
     FOnNameChange:TNameChangeEvent;
     procedure DoNewBounds; dynamic;
     {NB: CM_*** are defined in Controls.pas}
     procedure CMTextChanged(var Msg: TMessage); message CM_TEXTCHANGED;
     procedure CMFontChanged(var Msg: TMessage); message CM_FONTCHANGED;
     procedure CMColorChanged(var Msg: TMessage); message CM_COLORCHANGED;
     procedure SetBounds(ALeft, ATop, AWidth, AHeight: Integer); override;
     procedure Click; override;
     procedure SetDisabled(SetOn:Boolean);
     procedure SetDisabledColor(AColor:TColor);
    public
     constructor Create(AOwner: TComponent); override;
     {Properties used by labeled input controls}
     property Disabled:Boolean read FDisabled write SetDisabled; {use instead of Enabled}
     property OnNewBounds:TNotifyEvent read FOnNewBounds write FOnNewBounds;
     property OnNameChange:TNameChangeEvent read FOnNameChange write FOnNameChange;
    published
     property DisabledColor:TColor read FDisabledColor write SetDisabledColor
              default clBtnShadow;
  end;

  ELabeledControl = class(Exception)
    public
     constructor ShowError(ErrorId:Word; const AComponentName:TComponentName);
  end;

  TLblEdit = class(TEdit)
    protected
     FLabel:TLabel;
     FLabelPos:TLabelPos;
     FMargin:Integer;
     FAddColon,FRepositionLabel,FNamedBySelf:Boolean;
     procedure Loaded; override;
     procedure VisibleChanging; override;
     procedure WMEnable(var Msg:TMessage); message WM_ENABLE;
     procedure SetBounds(ALeft,ATop,AWidth,AHeight:Integer); override;
     procedure SetName(const NewName: TComponentName); override;
     procedure DestroyLabel;
     procedure InitLabel;
     procedure Notification(AComponent: TComponent; Operation: TOperation); override;
     procedure OnLabelNewBounds(Sender: TObject);
     procedure DrawLabel;
     procedure SetLabelCaption(const AString:string);
     function  GetLabelCaption:string;
     procedure SetLabelPos(APos:TLabelPos);
     procedure SetMargin(AMargin:Integer);
     procedure SetAddColon(SetOn:Boolean);
     procedure SetLabel(ALabel:TLabel);
    public
     constructor Create(AOwner: TComponent); override;
     destructor Destroy; override;
     function HasLabel:Boolean;
    published
     property LabelText:string read GetLabelCaption write SetLabelCaption stored False;
     property LabelPos:TLabelPos read FLabelPos write SetLabelPos default DefaultLabelPos;
     property LabelMargin:Integer read FMargin write SetMargin default DefaultMargin;
     property LabelColon:Boolean read FAddColon write SetAddColon default DefaultAddColon;
     property LabelComponent:TLabel read FLabel write SetLabel;
  end;

  TLblMemo = class(TEnhMemo)
    protected
     FLabel:TLabel;
     FLabelPos:TLabelPos;
     FMargin:Integer;
     FAddColon,FRepositionLabel,FNamedBySelf:Boolean;
     procedure Loaded; override;
     procedure VisibleChanging; override;
     procedure WMEnable(var Msg:TMessage); message WM_ENABLE;
     procedure SetBounds(ALeft,ATop,AWidth,AHeight:Integer); override;
     procedure SetName(const NewName: TComponentName); override;
     procedure DestroyLabel;
     procedure InitLabel;
     procedure Notification(AComponent: TComponent; Operation: TOperation); override;
     procedure OnLabelNewBounds(Sender: TObject);
     procedure DrawLabel;
     procedure SetLabelCaption(const AString:string);
     function  GetLabelCaption:string;
     procedure SetLabelPos(APos:TLabelPos);
     procedure SetMargin(AMargin:Integer);
     procedure SetAddColon(SetOn:Boolean);
     procedure SetLabel(ALabel:TLabel);
    public
     constructor Create(AOwner: TComponent); override;
     destructor Destroy; override;
     function HasLabel:Boolean;
    published
     property LabelText:string read GetLabelCaption write SetLabelCaption stored False;
     property LabelPos:TLabelPos read FLabelPos write SetLabelPos default DefaultLabelPos;
     property LabelMargin:Integer read FMargin write SetMargin default DefaultMargin;
     property LabelColon:Boolean read FAddColon write SetAddColon default DefaultAddColon;
     property LabelComponent:TLabel read FLabel write SetLabel;
  end;

  TLblComboBox = class(TComboBox)
    protected
     FLabel:TLabel;
     FLabelPos:TLabelPos;
     FMargin:Integer;
     FAddColon,FRepositionLabel,FNamedBySelf:Boolean;
     procedure Loaded; override;
     procedure VisibleChanging; override;
     procedure WMEnable(var Msg:TMessage); message WM_ENABLE;
     procedure SetBounds(ALeft,ATop,AWidth,AHeight:Integer); override;
     procedure SetName(const NewName: TComponentName); override;
     procedure DestroyLabel;
     procedure InitLabel;
     procedure Notification(AComponent: TComponent; Operation: TOperation); override;
     procedure OnLabelNewBounds(Sender: TObject);
     procedure DrawLabel;
     procedure SetLabelCaption(const AString:string);
     function  GetLabelCaption:string;
     procedure SetLabelPos(APos:TLabelPos);
     procedure SetMargin(AMargin:Integer);
     procedure SetAddColon(SetOn:Boolean);
     procedure SetLabel(ALabel:TLabel);
    public
     constructor Create(AOwner: TComponent); override;
     destructor Destroy; override;
     function HasLabel:Boolean;
    published
     property LabelText:string read GetLabelCaption write SetLabelCaption stored False;
     property LabelPos:TLabelPos read FLabelPos write SetLabelPos default DefaultLabelPos;
     property LabelMargin:Integer read FMargin write SetMargin default DefaultMargin;
     property LabelColon:Boolean read FAddColon write SetAddColon default DefaultAddColon;
     property LabelComponent:TLabel read FLabel write SetLabel;
  end;

  TLblMaskEdit = class(TMaskEdit)
    protected
     FLabel:TLabel;
     FLabelPos:TLabelPos;
     FMargin:Integer;
     FAddColon,FRepositionLabel,FNamedBySelf:Boolean;
     procedure Loaded; override;
     procedure VisibleChanging; override;
     procedure WMEnable(var Msg:TMessage); message WM_ENABLE;
     procedure SetBounds(ALeft,ATop,AWidth,AHeight:Integer); override;
     procedure SetName(const NewName: TComponentName); override;
     procedure DestroyLabel;
     procedure InitLabel;
     procedure Notification(AComponent: TComponent; Operation: TOperation); override;
     procedure OnLabelNewBounds(Sender: TObject);
     procedure DrawLabel;
     procedure SetLabelCaption(const AString:string);
     function  GetLabelCaption:string;
     procedure SetLabelPos(APos:TLabelPos);
     procedure SetMargin(AMargin:Integer);
     procedure SetAddColon(SetOn:Boolean);
     procedure SetLabel(ALabel:TLabel);
    public
     constructor Create(AOwner: TComponent); override;
     destructor Destroy; override;
     function HasLabel:Boolean;
    published
     property LabelText:string read GetLabelCaption write SetLabelCaption stored False;
     property LabelPos:TLabelPos read FLabelPos write SetLabelPos default DefaultLabelPos;
     property LabelMargin:Integer read FMargin write SetMargin default DefaultMargin;
     property LabelColon:Boolean read FAddColon write SetAddColon default DefaultAddColon;
     property LabelComponent:TLabel read FLabel write SetLabel;
  end;


procedure Register;

implementation

procedure TEnhMemo.SetTabStops(Tabs:array of Word);
var NumTabs:Word;
begin
   NumTabs := High(Tabs) + 1;
   Perform(EM_SetTabStops,NumTabs,Longint(@Tabs));
end;

procedure TEnhMemo.Copy2Printer;
var I:Integer; MemoText:TextFile;
begin
   AssignPrn(MemoText);
   try
      Rewrite(MemoText);
      Printer.Canvas.Font.Assign(Font);
      for I := 0 to Lines.Count - 1 do
         Writeln(MemoText,Lines[I]);
   finally
      CloseFile(MemoText);
   end;
end;

procedure TEnhMemo.ResetUndoFlag;
begin
   Perform(EM_EMPTYUNDOBUFFER,0,0);
end;

function TEnhMemo.CanUndo:Boolean;
begin
   Result := Perform(EM_CANUNDO,0,0) <> 0;
end;

procedure TEnhMemo.UndoEdit;
begin
   Perform(EM_UNDO,0,0);
end;

procedure TEnhMemo.CurrentCursorPos(var Row,Col:Integer);
begin
   Row := CursorRow;
   Col := SelStart - SendMessage(Handle,EM_LINEINDEX,Row-1,0)+1;
end;

function TEnhMemo.CursorRow:Integer;
begin
   Result := SendMessage(Handle,EM_LINEFROMCHAR,SelStart,0)+1;
end;

function TEnhMemo.CursorCol:Integer;
var Row,Col:Integer;
begin
   CurrentCursorPos(Row,Col);
   Result := Col;
end;

constructor TEnhLabel.Create(AOwner: TComponent);
begin
   inherited Create(AOwner);
   FDisabledColor := clBtnShadow;
   FKeepColor := clBlack;
end;

procedure TEnhLabel.DoNewBounds;
begin
   if Assigned(FOnNewBounds) then FOnNewBounds(Self);
end;

procedure TEnhLabel.CMTextChanged(var Msg: TMessage);
begin
   inherited;
   if not AutoSize then DoNewBounds
end;

procedure TEnhLabel.CMFontChanged(var Msg: TMessage);
begin
   inherited;
   if not AutoSize then DoNewBounds
end;

procedure TEnhLabel.CMColorChanged(var Msg: TMessage);
begin
   inherited;
   if (not FDisabled) and (FKeepColor <> Font.Color) then FKeepColor := Font.Color;
end;

procedure TEnhLabel.SetBounds(ALeft, ATop, AWidth, AHeight: Integer);
begin
   inherited SetBounds(ALeft,ATop,AWidth,AHeight);
   DoNewBounds
end;

procedure TEnhLabel.Click;
{Modifies the method so that the focus control
 receives the focus when the label is clicked}
begin
   if not FDisabled then
      begin
         Inherited Click;
         if FocusControl <> nil then
            FocusControl.SetFocus;
      end;
end;

procedure TEnhLabel.SetDisabled(SetOn:Boolean);
begin
   if FDisabled <> SetOn then
      begin
         FDisabled := SetOn;
         if FDisabled then
            begin
               FKeepColor := Font.Color;
               Font.Color := FDisabledColor;
            end
         else
            Font.Color := FKeepColor;
         invalidate;
      end;
end;

procedure TEnhLabel.SetDisabledColor(AColor:TColor);
begin
   if FDisabledColor <> AColor then
      begin
         FDisabledColor := AColor;
         if FDisabled then
            Font.Color := FDisabledColor;
      end;
end;

constructor ELabeledControl.ShowError(ErrorId:Word; const AComponentName:TComponentName);
begin
   case ErrorId of
      ElcDiffParent : Create('Cannot set label position. Label parent is different from '+
                             AComponentName+' parent.');
      ElcDiffOwner : Create('Cannot attach label to '+AComponentName+
                            '. Label and control must be on the same form!');
      else Create('Undefined exception (?)');
   end;
end;

constructor TLblEdit.Create(AOwner: TComponent);
begin
   inherited Create(AOwner);
   FRepositionLabel := True;
   {Set default property values}
   FLabelPos := DefaultLabelPos;
   FMargin := DefaultMargin;
   FAddColon := DefaultAddColon;
end;

destructor TLblEdit.Destroy;
begin
   DestroyLabel;
   inherited Destroy;
end;

procedure TLblEdit.DestroyLabel;
begin
   {If FLabel exists and form or Parent control is not being destroyed}
   if (FLabel <> nil) and
   not ((csDestroying in Owner.ComponentState) or
   ((FLabel.Parent <> nil) and (csDestroying in FLabel.Parent.ComponentState))) then
      begin
         {Remove label from parent list}
         if FLabel.Parent <> nil then
            FLabel.Parent.RemoveControl(FLabel);
         FLabel.Destroy; {This executes Notification method}
      end;
end;

procedure TLblEdit.WMEnable(var Msg:TMessage);
begin
   {This procedure is called whenever the value of the Enable property changes}
   if FLabel <> nil then
      begin
         if FLabel is TEnhLabel then
            TEnhLabel(FLabel).Disabled := not Boolean(Msg.wParam)
         else
            FLabel.Enabled := Boolean(Msg.wParam);
      end;
   inherited;
end;

procedure TLblEdit.SetBounds(ALeft,ATop,AWidth,AHeight:Integer);
begin
   inherited SetBounds(ALeft,ATop,AWidth,AHeight);
   if (FLabel <> nil) and not (csLoading in ComponentState) then
      DrawLabel;  {Redraw label whenever input component is moved or resized}
end;

procedure TLblEdit.SetName(const NewName: TComponentName);
var RenameLabel:Boolean;
begin
   if csLoading in ComponentState then
      inherited SetName(NewName)
   else
      begin
         RenameLabel := (FLabel <> nil) and (FLabel.Name = LabelPrefix+Name);
         inherited SetName(NewName);
         {Change label name if it is based on input component name}
         if RenameLabel then FLabel.Name := LabelPrefix+Name;
      end;
end;

procedure TLblEdit.InitLabel;
begin
   if FLabel = nil then
      begin
         FLabel := TEnhLabel.Create(Owner);
         {Use the same Parent as the input control}
         FLabel.Parent := Parent;
         FLabel.Name := LabelPrefix+Name;
         {Make the input control the focus control of the label}
         FLabel.FocusControl := Self;
         FLabel.Visible := False;
         {This notifies us whenever the label's position or size changes}
         TEnhLabel(FLabel).OnNewBounds := OnLabelNewBounds;
      end;
end;

procedure TLblEdit.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  {If our attached label is being destroyed, then set its pointer to nil - avoids GPFs}
  if (Operation = opRemove) and (AComponent = FLabel) then FLabel := nil;
end;

procedure TLblEdit.OnLabelNewBounds(Sender: TObject);
begin
   {If label has changed size or has moved and requires repositioning, then redraw it}
   if FRepositionLabel and (FLabelPos <> lpCustom) and
   not ((csLoading in ComponentState) or (csLoading in FLabel.ComponentState)) then
      DrawLabel;
end;

function TLblEdit.HasLabel:Boolean;
begin
   Result := (FLabel <> nil) and (FLabel.Caption <> '');
end;

procedure TLblEdit.Loaded;
begin
   inherited Loaded;
   {Form file has now been read and properties have been set;
    can now draw the label if there is one. This method avoids
    multiple redraws when each label related property is set}
   if HasLabel and FLabel.AutoSize then
      begin
         {This ensures the label is really the right size
          when screen resolution changes}
         FLabel.AutoSize := False;
         FLabel.AutoSize := True;
      end;
   DrawLabel;
   {Adjust the label's Disable property to its input control's Enabled property}
   {Note that this property should not be set at design time, otherwise the original
    label color will be lost if Enable is False}
   if (FLabel <> nil) and not (csDesigning in ComponentState) then
      if FLabel is TEnhLabel then
         TEnhLabel(FLabel).Disabled := not Enabled
      else
         FLabel.Enabled := Enabled; {unfortunately, setting false often hides standard labels}
end;

procedure TLblEdit.VisibleChanging;
{The redefined method makes sure the label is hidden or show
 together with the input control}
begin
   if FLabel <> nil then
      if Visible then
         FLabel.Visible := False
      else
         begin
            if not (csLoading in ComponentState) then DrawLabel;
            FLabel.Visible := True;
         end;
   inherited VisibleChanging;
end;

procedure TLblEdit.DrawLabel;
var LLeft,LTop:Integer;
begin
   if FLabel <> nil then
      begin
         case FLabelPos of
            lpTop : begin
                       LLeft := Left; LTop := Top - (FLabel.Height+FMargin);
                    end;
           lpLeft : begin
                       LLeft := Left - (FLabel.Width+FMargin);
                       LTop := Top + AdjLabelToControl;
                    end;
         lpBottom : begin
                       LLeft := Left; LTop := Top + Height + FMargin;
                    end;
          lpRight : begin
                       LLeft := Left + Width + FMargin;
                       LTop := Top + AdjLabelToControl;
                    end;
         end;
         {Only adjust label position if not custom position}
         if FLabelPos <> lpCustom then
            begin
               FRepositionLabel := False;
               FLabel.Setbounds(LLeft,LTop,FLabel.Width,FLabel.Height);
               FRepositionLabel := True;
            end;
         {Set same label visibility as the input control}
         FLabel.Visible := Visible;
      end;
end;

procedure TLblEdit.SetLabelCaption(const AString:string);
begin
   if not (csLoading in ComponentState) then {used for compatibility with earlier versions}
      begin
         if FLabel = nil then
            begin
               FRepositionLabel := False;
               {If label does not exist yet, then create it}
               InitLabel;
            end;
         FRepositionLabel := True;
         if FAddColon and (AString <> '') and (AString[Length(AString)] <> ':') then
            FLabel.Caption := AString + ':'
         else
            FLabel.Caption := AString;
      end;
end;

function TLblEdit.GetLabelCaption:string;
begin
   if HasLabel then
      begin
         Result := FLabel.Caption;
         if FAddColon then Delete(Result,Length(Result),1); {Removes the added colon}
      end
   else
      Result := '';
end;

procedure TLblEdit.SetLabelPos(APos:TLabelPos);
begin
   if csLoading in ComponentState then
      FLabelPos := APos
   else if FLabelPos <> APos then
      begin
         {Label must have same parent as input control if label
          position is other than lpCustom}
         if (FLabel <> nil) and (FLabel.Parent <> Parent) then
            raise ELabeledControl.ShowError(ElcDiffParent,Name);
         {Sets most appropriate margin at when label position property
          is changed at design time}
         if DefaultLabelPos = lpTop then
            begin
               if (FLabelPos = lpTop) and (APos <> lpBottom) and (FMargin = 2)
               and (csDesigning in ComponentState) then
                  FMargin := 8;
            end
         else
            begin
               if (FLabelPos = DefaultLabelPos) and (APos = lpTop) and (FMargin > 2)
               and (csDesigning in ComponentState) then
                  FMargin := 2;
            end;
         FLabelPos := APos;
         DrawLabel;
      end;
end;

procedure TLblEdit.SetMargin(AMargin:Integer);
begin
   if csLoading in ComponentState then
      FMargin := AMargin
   else if FMargin <> AMargin then
      begin
         FMargin := AMargin;
         DrawLabel;
      end;
end;

procedure TLblEdit.SetAddColon(SetOn:Boolean);
var LabelStr:string;
begin
   if csLoading in ComponentState then
      FAddColon := SetOn
   else if FAddColon <> SetOn then
      begin
         LabelStr := GetLabelCaption;
         FAddColon := SetOn;
         if LabelStr <> '' then
            SetLabelCaption(LabelStr);
      end;
end;

procedure TLblEdit.SetLabel(ALabel:TLabel);
{Assigns a nil or pre-existing label}
var AString:string;
begin
   if FLabel <> ALabel then
      if (ALabel = nil) or ((ALabel.FocusControl is ClassType)
      and (ALabel.FocusControl <> Self)) then
         {If ALabel is nil, then FLabel is not destroyed but detached from input control}
         FLabel := nil
      else if csLoading in ComponentState then
         begin
            FLabel := ALabel;
            if FLabel is TEnhLabel then
               TEnhLabel(FLabel).OnNewBounds := OnLabelNewBounds;
         end
      else
         begin
            {The attached label must be on the same form as the input control}
            if ALabel.Owner <> Owner then
               raise ELabeledControl.ShowError(ElcDiffOwner,Name);
            DestroyLabel;
            {Position is not changed so FLabelPos is set to lpCustom}
            FLabelPos := lpCustom;
            FLabel := ALabel;
            if FLabel.Name = '' then
               FLabel.Name := LabelPrefix+Name;
            FLabel.FocusControl := Self;
           {If the label is of type TEnhLabel then make use of its specific properties}
            if FLabel is TEnhLabel then
               TEnhLabel(FLabel).OnNewBounds := OnLabelNewBounds;
            {Check if new label caption has a colon and adjust colon property}
            AString := FLabel.Caption;
            LabelColon := (AString[Length(AString)] = ':');
         end;
end;

constructor TLblMemo.Create(AOwner: TComponent);
begin
   inherited Create(AOwner);
   FRepositionLabel := True;
   FLabelPos := DefaultLabelPos;
   FMargin := DefaultMargin;
   FAddColon := DefaultAddColon;
end;

destructor TLblMemo.Destroy;
begin
   DestroyLabel;
   inherited Destroy;
end;

procedure TLblMemo.DestroyLabel;
begin
   if (FLabel <> nil) and
   not ((csDestroying in Owner.ComponentState) or
   ((FLabel.Parent <> nil) and (csDestroying in FLabel.Parent.ComponentState))) then
      begin
         if FLabel.Parent <> nil then
            FLabel.Parent.RemoveControl(FLabel);
         FLabel.Destroy;
      end;
end;

procedure TLblMemo.WMEnable(var Msg:TMessage);
begin
   if FLabel <> nil then
      begin
         if FLabel is TEnhLabel then
            TEnhLabel(FLabel).Disabled := not Boolean(Msg.wParam)
         else
            FLabel.Enabled := Boolean(Msg.wParam);
      end;
   inherited;
end;

procedure TLblMemo.SetBounds(ALeft,ATop,AWidth,AHeight:Integer);
begin
   inherited SetBounds(ALeft,ATop,AWidth,AHeight);
   if (FLabel <> nil) and not (csLoading in ComponentState) then
      DrawLabel;
end;

procedure TLblMemo.SetName(const NewName: TComponentName);
var RenameLabel:Boolean;
begin
   if csLoading in ComponentState then
      inherited SetName(NewName)
   else
      begin
         RenameLabel := (FLabel <> nil) and (FLabel.Name = LabelPrefix+Name);
         inherited SetName(NewName);
         if RenameLabel then FLabel.Name := LabelPrefix+Name;
      end;
end;

procedure TLblMemo.InitLabel;
begin
   if FLabel = nil then
      begin
         FLabel := TEnhLabel.Create(Owner);
         FLabel.Parent := Parent;
         FLabel.Name := LabelPrefix+Name;
         FLabel.FocusControl := Self;
         FLabel.Visible := False;
         TEnhLabel(FLabel).OnNewBounds := OnLabelNewBounds;
      end;
end;

procedure TLblMemo.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (Operation = opRemove) and (AComponent = FLabel) then FLabel := nil;
end;

procedure TLblMemo.OnLabelNewBounds(Sender: TObject);
begin
   if FRepositionLabel and (FLabelPos <> lpCustom) and
   not ((csLoading in ComponentState) or (csLoading in FLabel.ComponentState)) then
      DrawLabel;
end;

function TLblMemo.HasLabel:Boolean;
begin
   Result := (FLabel <> nil) and (FLabel.Caption <> '');
end;

procedure TLblMemo.Loaded;
begin
   inherited Loaded;
   if HasLabel and FLabel.AutoSize then
      begin
         FLabel.AutoSize := False;
         FLabel.AutoSize := True;
      end;
   DrawLabel;
   if (FLabel <> nil) and not (csDesigning in ComponentState) then
      if FLabel is TEnhLabel then
         TEnhLabel(FLabel).Disabled := not Enabled
      else
         FLabel.Enabled := Enabled;
end;

procedure TLblMemo.VisibleChanging;
begin
   if FLabel <> nil then
      if Visible then
         FLabel.Visible := False
      else
         begin
            if not (csLoading in ComponentState) then DrawLabel;
            FLabel.Visible := True;
         end;
   inherited VisibleChanging;
end;

procedure TLblMemo.DrawLabel;
var LLeft,LTop:Integer;
begin
   if FLabel <> nil then
      begin
         case FLabelPos of
            lpTop : begin
                       LLeft := Left; LTop := Top - (FLabel.Height+FMargin);
                    end;
           lpLeft : begin
                       LLeft := Left - (FLabel.Width+FMargin);
                       LTop := Top + AdjLabelToControl;
                    end;
         lpBottom : begin
                       LLeft := Left; LTop := Top + Height + FMargin;
                    end;
          lpRight : begin
                       LLeft := Left + Width + FMargin;
                       LTop := Top + AdjLabelToControl;
                    end;
         end;
         if FLabelPos <> lpCustom then
            begin
               FRepositionLabel := False;
               FLabel.Setbounds(LLeft,LTop,FLabel.Width,FLabel.Height);
               FRepositionLabel := True;
            end;
         FLabel.Visible := Visible;
      end;
end;

procedure TLblMemo.SetLabelCaption(const AString:string);
begin
   if not (csLoading in ComponentState) then
      begin
         if FLabel = nil then
            begin
               FRepositionLabel := False;
               InitLabel;
            end;
         FRepositionLabel := True;
         if FAddColon and (AString <> '') and (AString[Length(AString)] <> ':') then
            FLabel.Caption := AString + ':'
         else
            FLabel.Caption := AString;
      end;
end;

function TLblMemo.GetLabelCaption:string;
begin
   if HasLabel then
      begin
         Result := FLabel.Caption;
         if FAddColon then Delete(Result,Length(Result),1);
      end
   else
      Result := '';
end;

procedure TLblMemo.SetLabelPos(APos:TLabelPos);
begin
   if csLoading in ComponentState then
      FLabelPos := APos
   else if FLabelPos <> APos then
      begin
         if (FLabel <> nil) and (FLabel.Parent <> Parent) then
            raise ELabeledControl.ShowError(ElcDiffParent,Name);
         if DefaultLabelPos = lpTop then
            begin
               if (FLabelPos = lpTop) and (APos <> lpBottom) and (FMargin = 2)
               and (csDesigning in ComponentState) then
                  FMargin := 8;
            end
         else
            begin
               if (FLabelPos = DefaultLabelPos) and (APos = lpTop) and (FMargin > 2)
               and (csDesigning in ComponentState) then
                  FMargin := 2;
            end;
         FLabelPos := APos;
         DrawLabel;
      end;
end;

procedure TLblMemo.SetMargin(AMargin:Integer);
begin
   if csLoading in ComponentState then
      FMargin := AMargin
   else if FMargin <> AMargin then
      begin
         FMargin := AMargin;
         DrawLabel;
      end;
end;

procedure TLblMemo.SetAddColon(SetOn:Boolean);
var LabelStr:string; 
begin
   if csLoading in ComponentState then
      FAddColon := SetOn
   else if FAddColon <> SetOn then
      begin
         LabelStr := GetLabelCaption;
         FAddColon := SetOn;
         if LabelStr <> '' then
            SetLabelCaption(LabelStr);
      end;
end;

procedure TLblMemo.SetLabel(ALabel:TLabel);
var AString:string;
begin
   if FLabel <> ALabel then
      if (ALabel = nil) or ((ALabel.FocusControl is ClassType)
      and (ALabel.FocusControl <> Self)) then
         FLabel := nil
      else if csLoading in ComponentState then
         begin
            FLabel := ALabel;
            if FLabel is TEnhLabel then
               TEnhLabel(FLabel).OnNewBounds := OnLabelNewBounds;
         end
      else
         begin
            if ALabel.Owner <> Owner then
               raise ELabeledControl.ShowError(ElcDiffOwner,Name);
            DestroyLabel;
            FLabelPos := lpCustom;
            FLabel := ALabel;
            if FLabel.Name = '' then
               FLabel.Name := LabelPrefix+Name;
            FLabel.FocusControl := Self;
            if FLabel is TEnhLabel then
               TEnhLabel(FLabel).OnNewBounds := OnLabelNewBounds;
            AString := FLabel.Caption;
            LabelColon := (AString[Length(AString)] = ':');
         end;
end;

constructor TLblComboBox.Create(AOwner: TComponent);
begin
   inherited Create(AOwner);
   FRepositionLabel := True;
   FLabelPos := DefaultLabelPos;
   FMargin := DefaultMargin;
   FAddColon := DefaultAddColon;
end;

destructor TLblComboBox.Destroy;
begin
   DestroyLabel;
   inherited Destroy;
end;

procedure TLblComboBox.DestroyLabel;
begin
   if (FLabel <> nil) and
   not ((csDestroying in Owner.ComponentState) or
   ((FLabel.Parent <> nil) and (csDestroying in FLabel.Parent.ComponentState))) then
      begin
         if FLabel.Parent <> nil then
            FLabel.Parent.RemoveControl(FLabel);
         FLabel.Destroy;
      end;
end;

procedure TLblComboBox.WMEnable(var Msg:TMessage);
begin
   if FLabel <> nil then
      begin
         if FLabel is TEnhLabel then
            TEnhLabel(FLabel).Disabled := not Boolean(Msg.wParam)
         else
            FLabel.Enabled := Boolean(Msg.wParam);
      end;
   inherited;
end;

procedure TLblComboBox.SetBounds(ALeft,ATop,AWidth,AHeight:Integer);
begin
   inherited SetBounds(ALeft,ATop,AWidth,AHeight);
   if (FLabel <> nil) and not (csLoading in ComponentState) then
      DrawLabel;
end;

procedure TLblComboBox.SetName(const NewName: TComponentName);
var RenameLabel:Boolean;
begin
   if csLoading in ComponentState then
      inherited SetName(NewName)
   else
      begin
         RenameLabel := (FLabel <> nil) and (FLabel.Name = LabelPrefix+Name);
         inherited SetName(NewName);
         if RenameLabel then FLabel.Name := LabelPrefix+Name;
      end;
end;

procedure TLblComboBox.InitLabel;
begin
   if FLabel = nil then
      begin
         FLabel := TEnhLabel.Create(Owner);
         FLabel.Parent := Parent;
         FLabel.Name := LabelPrefix+Name;
         FLabel.FocusControl := Self;
         FLabel.Visible := False;
         TEnhLabel(FLabel).OnNewBounds := OnLabelNewBounds;
      end;
end;

procedure TLblComboBox.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (Operation = opRemove) and (AComponent = FLabel) then FLabel := nil;
end;

procedure TLblComboBox.OnLabelNewBounds(Sender: TObject);
begin
   if FRepositionLabel and (FLabelPos <> lpCustom) and
   not ((csLoading in ComponentState) or (csLoading in FLabel.ComponentState)) then
      DrawLabel;
end;

function TLblComboBox.HasLabel:Boolean;
begin
   Result := (FLabel <> nil) and (FLabel.Caption <> '');
end;

procedure TLblComboBox.Loaded;
begin
   inherited Loaded;
   if HasLabel and FLabel.AutoSize then
      begin
         FLabel.AutoSize := False;
         FLabel.AutoSize := True;
      end;
   DrawLabel;
   if (FLabel <> nil) and not (csDesigning in ComponentState) then
      if FLabel is TEnhLabel then
         TEnhLabel(FLabel).Disabled := not Enabled
      else
         FLabel.Enabled := Enabled;
end;

procedure TLblComboBox.VisibleChanging;
begin
   if FLabel <> nil then
      if Visible then
         FLabel.Visible := False
      else
         begin
            if not (csLoading in ComponentState) then DrawLabel;
            FLabel.Visible := True;
         end;
   inherited VisibleChanging;
end;

procedure TLblComboBox.DrawLabel;
var LLeft,LTop:Integer;
begin
   if FLabel <> nil then
      begin
         case FLabelPos of
            lpTop : begin
                       LLeft := Left; LTop := Top - (FLabel.Height+FMargin);
                    end;
           lpLeft : begin
                       LLeft := Left - (FLabel.Width+FMargin);
                       LTop := Top + AdjLabelToControl;
                    end;
         lpBottom : begin
                       LLeft := Left; LTop := Top + Height + FMargin;
                    end;
          lpRight : begin
                       LLeft := Left + Width + FMargin;
                       LTop := Top + AdjLabelToControl;
                    end;
         end;
         if FLabelPos <> lpCustom then
            begin
               FRepositionLabel := False;
               FLabel.Setbounds(LLeft,LTop,FLabel.Width,FLabel.Height);
               FRepositionLabel := True;
            end;
         FLabel.Visible := Visible;
      end;
end;

procedure TLblComboBox.SetLabelCaption(const AString:string);
begin
   if not (csLoading in ComponentState) then
      begin
         if FLabel = nil then
            begin
               FRepositionLabel := False;
               InitLabel;
            end;
         FRepositionLabel := True;
         if FAddColon and (AString <> '') and (AString[Length(AString)] <> ':') then
            FLabel.Caption := AString + ':'
         else
            FLabel.Caption := AString;
      end;
end;

function TLblComboBox.GetLabelCaption:string;
begin
   if HasLabel then
      begin
         Result := FLabel.Caption;
         if FAddColon then Delete(Result,Length(Result),1);
      end
   else
      Result := '';
end;

procedure TLblComboBox.SetLabelPos(APos:TLabelPos);
begin
   if csLoading in ComponentState then
      FLabelPos := APos
   else if FLabelPos <> APos then
      begin
         if (FLabel <> nil) and (FLabel.Parent <> Parent) then
            raise ELabeledControl.ShowError(ElcDiffParent,Name);
         if DefaultLabelPos = lpTop then
            begin
               if (FLabelPos = lpTop) and (APos <> lpBottom) and (FMargin = 2)
               and (csDesigning in ComponentState) then
                  FMargin := 8;
            end
         else
            begin
               if (FLabelPos = DefaultLabelPos) and (APos = lpTop) and (FMargin > 2)
               and (csDesigning in ComponentState) then
                  FMargin := 2;
            end;
         FLabelPos := APos;
         DrawLabel;
      end;
end;

procedure TLblComboBox.SetMargin(AMargin:Integer);
begin
   if csLoading in ComponentState then
      FMargin := AMargin
   else if FMargin <> AMargin then
      begin
         FMargin := AMargin;
         DrawLabel;
      end;
end;

procedure TLblComboBox.SetAddColon(SetOn:Boolean);
var LabelStr:string; 
begin
   if csLoading in ComponentState then
      FAddColon := SetOn
   else if FAddColon <> SetOn then
      begin
         LabelStr := GetLabelCaption;
         FAddColon := SetOn;
         if LabelStr <> '' then
            SetLabelCaption(LabelStr);
      end;
end;

procedure TLblComboBox.SetLabel(ALabel:TLabel);
var AString:string;
begin
   if FLabel <> ALabel then
      if (ALabel = nil) or ((ALabel.FocusControl is ClassType)
      and (ALabel.FocusControl <> Self)) then
         FLabel := nil
      else if csLoading in ComponentState then
         begin
            FLabel := ALabel;
            if FLabel is TEnhLabel then
               TEnhLabel(FLabel).OnNewBounds := OnLabelNewBounds;
         end
      else
         begin
            if ALabel.Owner <> Owner then
               raise ELabeledControl.ShowError(ElcDiffOwner,Name);
            DestroyLabel;
            FLabelPos := lpCustom;
            FLabel := ALabel;
            if FLabel.Name = '' then
               FLabel.Name := LabelPrefix+Name;
            FLabel.FocusControl := Self;
            if FLabel is TEnhLabel then
               TEnhLabel(FLabel).OnNewBounds := OnLabelNewBounds;
            AString := FLabel.Caption;
            LabelColon := (AString[Length(AString)] = ':');
         end;
end;

constructor TLblMaskEdit.Create(AOwner: TComponent);
begin
   inherited Create(AOwner);
   FRepositionLabel := True;
   FLabelPos := DefaultLabelPos;
   FMargin := DefaultMargin;
   FAddColon := DefaultAddColon;
end;

destructor TLblMaskEdit.Destroy;
begin
   DestroyLabel;
   inherited Destroy;
end;

procedure TLblMaskEdit.DestroyLabel;
begin
   if (FLabel <> nil) and
   not ((csDestroying in Owner.ComponentState) or
   ((FLabel.Parent <> nil) and (csDestroying in FLabel.Parent.ComponentState))) then
      begin
         if FLabel.Parent <> nil then
            FLabel.Parent.RemoveControl(FLabel);
         FLabel.Destroy;
      end;
end;

procedure TLblMaskEdit.WMEnable(var Msg:TMessage);
begin
   if FLabel <> nil then
      begin
         if FLabel is TEnhLabel then
            TEnhLabel(FLabel).Disabled := not Boolean(Msg.wParam)
         else
            FLabel.Enabled := Boolean(Msg.wParam);
      end;
   inherited;
end;

procedure TLblMaskEdit.SetBounds(ALeft,ATop,AWidth,AHeight:Integer);
begin
   inherited SetBounds(ALeft,ATop,AWidth,AHeight);
   if (FLabel <> nil) and not (csLoading in ComponentState) then
      DrawLabel;
end;

procedure TLblMaskEdit.SetName(const NewName: TComponentName);
var RenameLabel:Boolean;
begin
   if csLoading in ComponentState then
      inherited SetName(NewName)
   else
      begin
         RenameLabel := (FLabel <> nil) and (FLabel.Name = LabelPrefix+Name);
         inherited SetName(NewName);
         if RenameLabel then FLabel.Name := LabelPrefix+Name;
      end;
end;

procedure TLblMaskEdit.InitLabel;
begin
   if FLabel = nil then
      begin
         FLabel := TEnhLabel.Create(Owner);
         FLabel.Parent := Parent;
         FLabel.Name := LabelPrefix+Name;
         FLabel.FocusControl := Self;
         FLabel.Visible := False;
         TEnhLabel(FLabel).OnNewBounds := OnLabelNewBounds;
      end;
end;

procedure TLblMaskEdit.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (Operation = opRemove) and (AComponent = FLabel) then FLabel := nil;
end;

procedure TLblMaskEdit.OnLabelNewBounds(Sender: TObject);
begin
   if FRepositionLabel and (FLabelPos <> lpCustom) and
   not ((csLoading in ComponentState) or (csLoading in FLabel.ComponentState)) then
      DrawLabel;
end;

function TLblMaskEdit.HasLabel:Boolean;
begin
   Result := (FLabel <> nil) and (FLabel.Caption <> '');
end;

procedure TLblMaskEdit.Loaded;
begin
   inherited Loaded;
   if HasLabel and FLabel.AutoSize then
      begin
         FLabel.AutoSize := False;
         FLabel.AutoSize := True;
      end;
   DrawLabel;
   if (FLabel <> nil) and not (csDesigning in ComponentState) then
      if FLabel is TEnhLabel then
         TEnhLabel(FLabel).Disabled := not Enabled
      else
         FLabel.Enabled := Enabled;
end;

procedure TLblMaskEdit.VisibleChanging;
begin
   if FLabel <> nil then
      if Visible then
         FLabel.Visible := False
      else
         begin
            if not (csLoading in ComponentState) then DrawLabel;
            FLabel.Visible := True;
         end;
   inherited VisibleChanging;
end;

procedure TLblMaskEdit.DrawLabel;
var LLeft,LTop:Integer;
begin
   if FLabel <> nil then
      begin
         case FLabelPos of
            lpTop : begin
                       LLeft := Left; LTop := Top - (FLabel.Height+FMargin);
                    end;
           lpLeft : begin
                       LLeft := Left - (FLabel.Width+FMargin);
                       LTop := Top + AdjLabelToControl;
                    end;
         lpBottom : begin
                       LLeft := Left; LTop := Top + Height + FMargin;
                    end;
          lpRight : begin
                       LLeft := Left + Width + FMargin;
                       LTop := Top + AdjLabelToControl;
                    end;
         end;
         if FLabelPos <> lpCustom then
            begin
               FRepositionLabel := False;
               FLabel.Setbounds(LLeft,LTop,FLabel.Width,FLabel.Height);
               FRepositionLabel := True;
            end;
         FLabel.Visible := Visible;
      end;
end;

procedure TLblMaskEdit.SetLabelCaption(const AString:string);
begin
   if not (csLoading in ComponentState) then
      begin
         if FLabel = nil then
            begin
               FRepositionLabel := False;
               InitLabel;
            end;
         FRepositionLabel := True;
         if FAddColon and (AString <> '') and (AString[Length(AString)] <> ':') then
            FLabel.Caption := AString + ':'
         else
            FLabel.Caption := AString;
      end;
end;

function TLblMaskEdit.GetLabelCaption:string;
begin
   if HasLabel then
      begin
         Result := FLabel.Caption;
         if FAddColon then Delete(Result,Length(Result),1);
      end
   else
      Result := '';
end;

procedure TLblMaskEdit.SetLabelPos(APos:TLabelPos);
begin
   if csLoading in ComponentState then
      FLabelPos := APos
   else if FLabelPos <> APos then
      begin
         if (FLabel <> nil) and (FLabel.Parent <> Parent) then
            raise ELabeledControl.ShowError(ElcDiffParent,Name);
         if DefaultLabelPos = lpTop then
            begin
               if (FLabelPos = lpTop) and (APos <> lpBottom) and (FMargin = 2)
               and (csDesigning in ComponentState) then
                  FMargin := 8;
            end
         else
            begin
               if (FLabelPos = DefaultLabelPos) and (APos = lpTop) and (FMargin > 2)
               and (csDesigning in ComponentState) then
                  FMargin := 2;
            end;
         FLabelPos := APos;
         DrawLabel;
      end;
end;

procedure TLblMaskEdit.SetMargin(AMargin:Integer);
begin
   if csLoading in ComponentState then
      FMargin := AMargin
   else if FMargin <> AMargin then
      begin
         FMargin := AMargin;
         DrawLabel;
      end;
end;

procedure TLblMaskEdit.SetAddColon(SetOn:Boolean);
var LabelStr:string; 
begin
   if csLoading in ComponentState then
      FAddColon := SetOn
   else if FAddColon <> SetOn then
      begin
         LabelStr := GetLabelCaption;
         FAddColon := SetOn;
         if LabelStr <> '' then
            SetLabelCaption(LabelStr);
      end;
end;

procedure TLblMaskEdit.SetLabel(ALabel:TLabel);
var AString:string;
begin
   if FLabel <> ALabel then
      if (ALabel = nil) or ((ALabel.FocusControl is ClassType)
      and (ALabel.FocusControl <> Self)) then
         FLabel := nil
      else if csLoading in ComponentState then
         begin
            FLabel := ALabel;
            if FLabel is TEnhLabel then
               TEnhLabel(FLabel).OnNewBounds := OnLabelNewBounds;
         end
      else
         begin
            if ALabel.Owner <> Owner then
               raise ELabeledControl.ShowError(ElcDiffOwner,Name);
            DestroyLabel;
            FLabelPos := lpCustom;
            FLabel := ALabel;
            if FLabel.Name = '' then
               FLabel.Name := LabelPrefix+Name;
            FLabel.FocusControl := Self;
            if FLabel is TEnhLabel then
               TEnhLabel(FLabel).OnNewBounds := OnLabelNewBounds;
            AString := FLabel.Caption;
            LabelColon := (AString[Length(AString)] = ':');
         end;
end;


procedure Register;
begin
   RegisterComponents('LblIC',[TEnhMemo,TEnhLabel,TLblEdit,TLblMemo,
                                TLblComboBox,TLblMaskEdit]);
end;


end.
